(ns com.dreamsoft.calculator.ast
	"Ease the creation and the incremental building of the abstract syntax tree,
	 abbreviated AST in the rest of this namespace"
	(:require [clojure.zip :as zip]
		      [com.dreamsoft.calculator.token :as token]))

(defn- zipper?
  "Test whether or not a value is a zipper"
  [x]
  (let [in-metadata? #(contains? (meta x) %)]
  	(every? in-metadata?  [:zip/branch? :zip/children :zip/make-node])))

(defn- append-child
  "Create a new AST with the token as the last child and focus on it"
  [loc0 token]
	{:pre [(zipper? loc0) (string? token)]}
  (-> (zip/append-child loc0 token)
      zip/down))
	; same as (zip/down (zip/append-child loc token) ast)

(defn- append-sibling
  "Create a new AST with the token as the last sibling and focus on it"
  [loc1 token]
	{:pre [(zipper? loc1) (string? token)]}
  (-> loc1 zip/up (zip/append-child token) zip/down))

; a function with a "?" at the end of his name is a predicat (returns `true` or `false`)
(defn- has-enough-children?
  "Check whether an AST's node is missing children"
  [loc2]
	{:pre [(zipper? loc2)]}
	; `let` defines local "variables"
  (let [token (zip/node loc2)
        children (zip/children loc2)]
    (>= (count children) (get token/children-needed token))))

(defn- contains-dot?
	"Check where the current number contains a dot"
	[loc3]
	{:pre [(zipper? loc3)]}
	(loop [loc3 loc3]
	  (let [current (zip/node loc3)]
		(cond
		 (token/dot? current)
		  true
		 (nil? (zip/down loc3))
		  false
		 :else
		  (recur (zip/down loc3))))))

(defn- ignore-token
  "Ignore a token and log the current AST state"
  [loc4 token]
	{:pre [(zipper? loc4) (string? token)]}
	; `do` let you "chains operations" only for their side effects
	; it returns only the result of the last "operation"
  (let [parent (zip/up loc4)
  	    children (-> loc4 zip/up zip/children)]
  	(do
  	  (prn (str "invalid operator '" token "' inserted under '" parent "' with siblings '" children "'"))
      loc4)))

(defn- make-parent
  "Create a new AST with the specified token as parent
   and the old AST as first child"
  [loc5 x]
	{:pre [(zipper? loc5) (string? x)]}
	(zip/edit loc5
		        (fn [children]
							(conj [x] children))))
	; `(fn [arguments ...] ...)` defines an anonymous function

(defn- parent
  "Grab the parent node"
	[loc6]
	{:pre [(zipper? loc6)]}
  (-> loc6 zip/up zip/node))

(defn singleton
  "Create a new single token AST"
  [val]
  {:pre [(string? val)]}
  (zip/vector-zip val))

(defn- top?
  "Check whether a AST node is the top node"
  [loc7]
  {:pre [(zipper? loc7)]}
  (-> loc7 zip/up nil?))

(defn- closest-parent-match
  "Create a new AST with the closest parent matching the predicate focused on"
  [loc8 pred]
  {:pre [(zipper? loc8)]}
  ; `loop` defines local variables and their first values
  ; `recur` "redo" all operations from the `loop` but with different values
  (loop [current loc8]
	(cond
	 (top? (zip/node current))
	  loc8
	 (pred (zip/node current))
	  current
	 :else
	  (recur (zip/up current)))))

(defn- insert-digit [loc9 digit]
	{:pre [(zipper? loc9) (string? digit)]}
	(let [current (zip/node loc9)]
	  (cond
	   (or (token/operator? current)
		   (token/open-parenthesis? current))
		(if (has-enough-children? current)
		  (ignore-token loc9 token/dot)
		  (append-child loc9 digit))
	   (token/digit? (first current))
		(zip/edit loc9 token/append-digits digit))))

(defn- insert-dot [loc10]
	{:pre [(zipper? loc10)]}
	(let [current (zip/node loc10)]
	  (cond
	   (or (token/operator? current)
		   (token/open-parenthesis? current))
		  (if (has-enough-children? current)
			(ignore-token loc10 token/dot)
			(append-child loc10 [token/dot ["0"]]))
		 (token/digit? (first current))
		  (if (contains-dot? current)
			loc10
			(zip/edit loc10 token/append-digits token/dot)))))

(defn- insert-open-parenthesis
	[loc11]
	{:pre [(zipper? loc11)]}
  (let [current (zip/node loc11)]
		; `cond` take "tests" and operations to do paires
		; if a "test" passes the corresponding operations are done and `cond` is exited
		; all "test"-operations paires that come after a passed "test" are ignored
		; => it works like `switch` in C language but each `case` has is own "test"
    (cond
     (token/operator? current)
      (if (has-enough-children? loc11)
        (ignore-token loc11 token/open-parenthesis)
        (append-child loc11 token/open-parenthesis)
     (token/operator? (parent loc11))
      (if (has-enough-children? (parent loc11))
        (ignore-token loc11 token/open-parenthesis)
        (append-sibling loc11 token/open-parenthesis))))))

(defn- insert-operator [loc12 operator]
	{:pre [(zipper? loc12) (token/operator? operator)]}
  (let [current (zip/node loc12)]
    (cond
     (token/operator? current)
      (zip/replace loc12 operator)
     (token/open-parenthesis? current)
      (append-child loc12 operator)
     (token/digit? (first current))
      (if (top? loc12)
        (make-parent loc12 operator)
        (ignore-token loc12 operator)))))

(defn insert-token
  "Create a new AST with the token inserted"
  [loc13 x]
  {:pre [(zipper? loc13) (string? x)]}
  (cond
   (token/open-parenthesis? x)
	(insert-open-parenthesis loc13)
   (token/close-parenthesis? x)
	(closest-parent-match loc13 token/open-parenthesis?)
   (token/operator? x)
	(insert-operator loc13 x)
   (token/digit? x)
	(insert-digit loc13 x)
   (token/dot? x)
	(insert-dot loc13))
   :else
	(ignore-token loc13 x))
