(ns com.dreamsoft.calculator.eval
  "Utilities to reduce the calculator AST as much as possible
   by evaluating it and getting the result"
  (:require [clojure.string :refer [join]]
  	  		[clojure.walk :as walk]
  	        [com.dreamsoft.calculator.bignum :as bignum]
  	        [com.dreamsoft.calculator.token :as token]))

(def ^:private operators-mapping
  "Mapping between tokens and evaluating functions"
  {token/plus bignum/plus
   token/minus bignum/minus
   token/times bignum/times
   token/divided bignum/divided})

(defn- eval-number
  "Specific function to evaluate numbers"
  [node]
  {:pre [(vector? node) (token/only-digits? node)]}
  (loop [acc []
  	     [x & xs] node]
    (if (empty? xs)
      (-> (conj acc x) join bignum/bignum)
      (recur (conj acc x) xs))))

;"eval" function predeclaration because of cyclic reference with "eval-operator" and "eval-parenthesis"
(declare eval)

(defn- eval-operator
  "Specific function to evaluate operators"
  [node]
	{:pre [(vector? node) (token/operator? (first node))]}
  (let [operator (get operators-mapping (first node))]
    (apply operator (map eval (rest node)))))

(defn- eval-parenthesis
  "Specific function to evaluate parenthesis"
  [node]
  {:pre [(vector? node)]}
  (eval (rest node)))

(defn eval
  "Main generic function to evaluate the AST by 'dispatching' on cases"
  [node]
  {:pre [(vector? node)]}
  (let [operation (first node)]
  	(cond
  	 (token/digit? operation)
  	  (eval-number node)
  	 (token/open-parenthesis? operation)
  	  (eval-parenthesis node)
  	 (token/operator? operation)
  	  (eval-operator node))))
