(ns com.dreamsoft.calculator.ui
	"Components for the user interface (UI) of the calculator"
	(:require [reagent.core :as rg]
		        [com.dreamsoft.calculator.eval :as eval]))

;; Reagent's components "domain specific language (DSL)"
; In reagent you define your UI components in a special DSL called 'Hiccups'.
; Hiccups simply uses ClojureScript data structures in a restricted way and
; give them a meaning for representing HTML and XML. Because it is just basic
; ClojureScript data strucutures, Hiccups is easy to read and can be easily
; manipulated and reusable by anobody for any library for any use case.
;
; The notation is very simple: it consist of nested 'vectors' (or arrays) where
; the first element is the node tag name, the second is a 'hash-map' (or
; associative table) which contains the attributes of the node and the following
; elements are the children of the node which are either strings or nodes in a
; similar way as the parent node. For example, a sample HTML document in Hiccups
; and then in plain HTML:
; 	[:html {}
;			[:head {}
;				[:meta {:charset "utf-8"}]
;				[:title {} "Title of the page"]
;			[:body {}
;				[:h1 {} "Title on the page"]
;				[:button {:on-click function-to-call-when-clicked}
;					"Click this button"]]]]
;
;	<html>
;		<head>
;			<meta charset="utf-8" />
;			<title>Title of the page</title>
;		</head>
;		<body>
;			<h1>Title on the page</h1>
;			<button onclick="functionToCallWhenClicked()">Click this button</button>
;		</body>
;	</html>

(defn button
  "A button widget which call the callback with the symbol when pressed"
  [symbol callback]
	[:button {:on-click #(callback symbol)
	          :style {:font-size "30px"
	                  :height "200px"
	                  :width "200px"}}
		(str symbol)])

(defn- display
  "A raw value display. Used a the screen of the calculator."
  [value]
	(do
		(.log js/console (str @value))
  [:p {:style {:background-color "black"
	           :color "white"
	           :font-family "monospace"
	           :font-size "40px"
	           :padding "20px"
	           :text-align "right"}}
    (str (eval/eval @value))]))

(defn- grid
  "A grid which display the items left to right and left to bottom.
   The grid can be scrolled on overflow."
  [items]
  (into []
        (concat
		  [:div {:style {:display "flex"
	                     :flex-direction "row"
	                     :flex-wrap "wrap"
	                     :overflow-y "auto"}}]
	      items)))

(defn page [value items]
  "A page which contains the calculator diplay and buttons"
	[:div {:style {:align-items "stretch"
	               :display "flex"
	               :flex "1"
	               :flex-direction "column"
	               :overflow "hidden"
	               :width "100%"}}
	    [display value]
	    [grid items]])
